import { CompilerCallbacks } from "./compiler-callbacks.js";
/**
 * Abstract interface for compiler error and warning messages
 */
export interface CompilerEvent {
    filename?: string;
    line?: number;
    code: number;
    message: string;
    /**
     * detailed Markdown-formatted description of the error including
     * references to documentation, remediation options.
     */
    detail?: string;
    /**
     * an internal error occurred that should be captured with a stack trace
     * e.g. to the Keyman sentry instance by kmc
     */
    exceptionVar?: any;
}
export declare enum CompilerErrorSeverity {
    Debug = 0,// log everything including internal debug
    Verbose = 1048576,// log everything, except debug
    Info = 2097152,// Informational, not necessarily a problem
    Hint = 3145728,// Something the user might want to be aware of
    Warn = 4194304,// Warning: Not great, but we can keep going.
    Error = 5242880,// Severe error where we can't continue
    Fatal = 6291456
}
export declare const CompilerErrorSeverityValues: CompilerErrorSeverity[];
/**
 * Mask values for mapping compiler errors
 */
export declare enum CompilerErrorMask {
    Severity = 15728640,// includes reserved bits, 16 possible severity levels
    Error = 1048575,// error | namespace
    Namespace = 1044480,// 256 possible namespaces
    BaseError = 4095,// error code, 2,048 possible error codes per namespace
    Reserved = 4278190080
}
export declare class CompilerError {
    static severity(code: number): CompilerErrorSeverity;
    static error(code: number): number;
    static baseError(code: number): number;
    static namespace(code: number): CompilerErrorNamespace;
    static formatSeverity(code: number): string;
    /** true if events has at least one message of the atLeast severity */
    static hasSeverity(events: CompilerEvent[], atLeast: CompilerErrorSeverity): boolean;
    /** true if events has at least one Error or worse */
    static hasError(events: CompilerEvent[]): boolean;
    /**
     * Format an error code number. The error code number does not include
     * the severity mask, as this is reported in text form separately; see
     * `severityName`.
     * @example
     *
     * The following call returns `KM03004`
     * ```
     *   formatCode(CompilerMessage.ERROR_InvalidDisplayMapFile)
     * ```
     */
    static formatCode(code: number): string;
    /**
     * Formats an event filename for an error report,
     * stripping off path component
     * @param filename
     * @returns
     */
    static formatFilename(filename: string, options?: {
        fullPath?: boolean;
        forwardSlashes?: boolean;
    }): string;
    /**
     * Formats an event line for an error report
     * @param line
     * @returns
     */
    static formatLine(line: number): string;
    /**
     * Formats an event message for an error report
     * @param message
     * @returns
     */
    static formatMessage(message: string): string;
    /**
     * Formats a compiler message, without coloring; an ANSI color version is
     * implemented in NodeCompilerCallbacks.
     * @param event event or array of events
     */
    static formatEvent(event: CompilerEvent | CompilerEvent[]): string;
    /**
     * @param e Error-like
     */
    static exceptionToString(e?: any): string;
    /**
     * Returns the corresponding error severity value from a partial name match,
     * e.g. 'inf' returns CompilerErrorSeverity.Info, or returns null if not found
     * @param name
     * @returns
     */
    static severityNameToValue(name: string): CompilerErrorSeverity;
}
/** @deprecated use `CompilerError.severity` instead */
export declare function compilerErrorSeverity(code: number): CompilerErrorSeverity;
/** @deprecated use `CompilerError.formatSeverity` instead */
export declare function compilerErrorSeverityName(code: number): string;
/** @deprecated use `CompilerError.formatCode` instead */
export declare function compilerErrorFormatCode(code: number): string;
/** @deprecated use `CompilerError.formatEvent` instead */
export declare function compilerEventFormat(e: CompilerEvent | CompilerEvent[]): string;
/**
 * Defines the error code ranges for various compilers. Once defined, these
 * ranges must not be changed as external modules may depend on specific error
 * codes. Individual errors are defined at a compiler level, for example,
 * kmc-ldml/src/compiler/messages.ts.
 *
 * kmc defines a mapping between each namespace and the corresponding compiler's
 * error reporting class in kmc/src/messages/messageNamespaces.ts
 */
export declare enum CompilerErrorNamespace {
    /**
     * kmc-ldml errors between 0x0000…0x0FFF
     */
    LdmlKeyboardCompiler = 0,
    /**
     * common/web/types errors between 0x1000…0x1FFF
     */
    CommonTypes = 4096,
    /**
     * kmc-kmn errors between 0x2000…0x2FFF; these map to
     * the base codes found in kmn_compiler_errors.h, exclusive severity flags
     */
    KmnCompiler = 8192,
    /**
     * kmc-model errors between 0x3000…0x3FFF
     */
    ModelCompiler = 12288,
    /**
     * kmc-package errors between 0x4000…0x4FFF
     */
    PackageCompiler = 16384,
    /**
     * kmc and related infrastructure errors between 0x5000…0x5FFF
     */
    Infrastructure = 20480,
    /**
     * kmc-analyze 0x6000…0x6FFF
     */
    Analyzer = 24576,
    /**
     * kmc-kmn/kmw-compiler errors between 0x7000…0x7FFF; note that some errors
     * generated by kmc-kmn/kmw-compiler are from kmc-kmn namespace for legacy
     * reasons
     */
    KmwCompiler = 28672,
    /**
     * kmc-model-info 0x8000…0x8FFF
     */
    ModelInfoCompiler = 32768,
    /**
     * kmc-keyboard-info 0x9000…0x9FFF
     */
    KeyboardInfoCompiler = 36864,
    /**
     * kmc-generate 0xA000…0xAFFF
     */
    Generator = 40960,
    /**
     * kmc-copy 0xB000…0xBFFF
     */
    Copier = 45056
}
type CompilerErrorSeverityOverride = CompilerErrorSeverity | 'disable';
export interface CompilerMessageOverrideMap {
    [code: number]: CompilerErrorSeverityOverride;
}
export interface CompilerMessageOverride {
    code: number;
    level: CompilerErrorSeverityOverride;
}
export interface CompilerCallbackOptions extends CompilerBaseOptions {
    compilerWarningsAsErrors?: boolean;
    messageOverrides?: CompilerMessageOverrideMap;
}
export interface KeymanCompilerArtifact {
    data: Uint8Array;
    filename: string;
}
export type KeymanCompilerArtifactOptional = KeymanCompilerArtifact | undefined;
export interface KeymanCompilerArtifacts {
    readonly [type: string]: KeymanCompilerArtifactOptional;
}
export interface KeymanCompilerResult {
    artifacts: KeymanCompilerArtifacts;
}
export interface KeymanCompiler {
    init(callbacks: CompilerCallbacks, options: CompilerOptions): Promise<boolean>;
    /**
     * Run the compiler, and save the result in memory arrays. Note that while
     * `outputFilename` is provided here, the output file is not written to in
     * this function.
     * @param inputFilename
     * @param outputFilename The intended output filename, optional, if missing,
     *                       calculated from inputFilename
     * @param data
     */
    run(inputFilename: string, outputFilename?: string): Promise<KeymanCompilerResult>;
    /**
     * Writes the compiled output files to disk
     * @param artifacts
     */
    write(artifacts: KeymanCompilerArtifacts): Promise<boolean>;
}
/**
 * Abstract interface for compiler options
 */
export interface CompilerBaseOptions {
    /**
     * Reporting level to console, used by NodeCompilerCallbacks (not used in compiler modules;
     * all messages are still reported to the internal log)
     */
    logLevel?: CompilerLogLevel;
    /**
     * Format of output for log to console
     */
    logFormat?: CompilerLogFormat;
    /**
     * Colorize log output, default is detected from console
     */
    color?: boolean;
}
export interface CompilerOptions extends CompilerBaseOptions {
    /**
     * Add metadata about the compiler version to .kmx file when compiling
     */
    shouldAddCompilerVersion?: boolean;
    /**
     * Add debug information to the .kmx file when compiling
     */
    saveDebug?: boolean;
    /**
     * Upgrade any warnings produced in the compile to errors
     */
    compilerWarningsAsErrors?: boolean;
    /**
     * Emit warnings if deprecated code is encountered
     */
    warnDeprecatedCode?: boolean;
    /**
     * Check filename conventions in packages
     */
    checkFilenameConventions?: boolean;
}
export declare const defaultCompilerOptions: CompilerOptions;
/**
 * Convenience function for constructing CompilerEvents
 * @param code     Unique numeric value of the event
 * @param message  A short description of the error presented to the user
 * @param detail   Detailed Markdown-formatted description of the error
 *                 including references to documentation, remediation options.
 * @returns
 */
export declare const CompilerMessageSpec: (code: number, message: string, detail?: string) => CompilerEvent;
/**
 * Remove initial whitespace from compiler detail messages, to enable
 * indented formatting of message detail strings inside the message
 * definitions
 * @param event
 * @returns dedented event detail
 */
export declare function dedentCompilerMessageDetail(event: CompilerEvent): string;
export declare const CompilerMessageDef: (param: any) => string;
export declare const CompilerMessageSpecWithException: (code: number, message: string, exceptionVar: any, detail?: string) => CompilerEvent;
/**
 * Compiler logging level and correspondence to severity
 */
export declare const ALL_COMPILER_LOG_LEVELS: readonly ["silent", "error", "warn", "hint", "info", "verbose", "debug"];
type CompilerLogLevelTuple = typeof ALL_COMPILER_LOG_LEVELS;
export type CompilerLogLevel = CompilerLogLevelTuple[number];
export declare const compilerLogLevelToSeverity: {
    [index in CompilerLogLevel]: number;
};
export declare const ALL_COMPILER_LOG_FORMATS: readonly ["tsv", "formatted"];
type CompilerLogFormatTuple = typeof ALL_COMPILER_LOG_FORMATS;
export type CompilerLogFormat = CompilerLogFormatTuple[number];
export {};
//# sourceMappingURL=compiler-interfaces.d.ts.map