import * as r from 'restructure';
import { constants } from '@keymanapp/ldml-keyboard-constants';
import { build_sect } from './build-sect.js';
import { build_disp } from './build-disp.js';
import { build_elem } from './build-elem.js';
import { build_keys } from './build-keys.js';
import { build_layr } from './build-layr.js';
import { build_list } from './build-list.js';
import { build_loca } from './build-loca.js';
import { build_meta } from './build-meta.js';
import { build_strs } from './build-strs.js';
import { build_tran } from './build-tran.js';
import { build_uset } from './build-uset.js';
import { build_vars } from './build-vars.js';
export default class KMXPlusBuilder {
    file;
    //private writeDebug: boolean;
    sect = {};
    constructor(file, _writeDebug) {
        this.file = file;
        //this.writeDebug = _writeDebug;
    }
    compile() {
        const fileSize = this.build();
        const file = new Uint8Array(fileSize);
        this.emitSection(file, this.file.COMP_PLUS_SECT, this.sect.sect);
        // Keep the rest of these in order.
        this.emitSection(file, this.file.COMP_PLUS_BKSP, this.sect.bksp);
        this.emitSection(file, this.file.COMP_PLUS_DISP, this.sect.disp);
        this.emitSection(file, this.file.COMP_PLUS_ELEM, this.sect.elem);
        this.emitElements(file);
        this.emitSection(file, this.file.COMP_PLUS_KEYS, this.sect.keys);
        this.emitSection(file, this.file.COMP_PLUS_LAYR, this.sect.layr);
        this.emitSection(file, this.file.COMP_PLUS_LIST, this.sect.list);
        this.emitSection(file, this.file.COMP_PLUS_LOCA, this.sect.loca);
        this.emitSection(file, this.file.COMP_PLUS_META, this.sect.meta);
        this.emitSection(file, this.file.COMP_PLUS_STRS, this.sect.strs);
        this.emitStrings(file);
        this.emitSection(file, this.file.COMP_PLUS_TRAN, this.sect.tran);
        this.emitSection(file, this.file.COMP_PLUS_USET, this.sect.uset);
        this.emitSection(file, this.file.COMP_PLUS_VARS, this.sect.vars);
        return file;
    }
    build() {
        // Required sections: sect, strs, loca, meta
        // We must prepare the strs, list, and elem sections early so that other sections can
        // reference them. However, they will be emitted in alpha order.
        this.sect.strs = build_strs(this.file.kmxplus.strs);
        this.sect.list = build_list(this.file.kmxplus.list, this.sect.strs);
        this.sect.uset = build_uset(this.file.kmxplus, this.sect.strs);
        this.sect.elem = build_elem(this.file.kmxplus.elem, this.sect.strs, this.sect.uset);
        const build_bksp = build_tran;
        this.sect.bksp = build_bksp(this.file.kmxplus.bksp, this.sect.strs, this.sect.elem);
        this.sect.disp = build_disp(this.file.kmxplus, this.sect.strs);
        this.sect.keys = build_keys(this.file.kmxplus, this.sect.strs, this.sect.list);
        this.sect.layr = build_layr(this.file.kmxplus, this.sect.strs, this.sect.list);
        this.sect.loca = build_loca(this.file.kmxplus, this.sect.strs);
        this.sect.meta = build_meta(this.file.kmxplus, this.sect.strs);
        this.sect.tran = build_tran(this.file.kmxplus.tran, this.sect.strs, this.sect.elem);
        this.sect.uset = build_uset(this.file.kmxplus, this.sect.strs);
        this.sect.vars = build_vars(this.file.kmxplus, this.sect.strs, this.sect.elem, this.sect.list);
        // Finalize the sect (index) section
        this.sect.sect = build_sect();
        this.finalize_sect(); // must be done last
        return this.sect.sect.total;
    }
    finalize_sect() {
        // 'sect' section
        this.sect.sect.count = 0;
        Object.keys(constants.section).forEach((sectstr) => {
            const sect = constants.section[sectstr];
            if (this.sect[sect] && sect !== 'sect') {
                this.sect.sect.count++;
            }
        });
        this.sect.sect.size = constants.length_sect + constants.length_sect_item * this.sect.sect.count;
        let offset = this.sect.sect.size;
        // Note: in order! Everyone's here except 'sect' which is at offset 0
        offset = this.finalize_sect_item(this.sect.bksp, offset);
        offset = this.finalize_sect_item(this.sect.disp, offset);
        offset = this.finalize_sect_item(this.sect.elem, offset);
        offset = this.finalize_sect_item(this.sect.keys, offset);
        offset = this.finalize_sect_item(this.sect.layr, offset);
        offset = this.finalize_sect_item(this.sect.list, offset);
        offset = this.finalize_sect_item(this.sect.loca, offset);
        offset = this.finalize_sect_item(this.sect.meta, offset);
        offset = this.finalize_sect_item(this.sect.strs, offset);
        offset = this.finalize_sect_item(this.sect.tran, offset);
        offset = this.finalize_sect_item(this.sect.uset, offset);
        offset = this.finalize_sect_item(this.sect.vars, offset);
        this.sect.sect.total = offset;
    }
    finalize_sect_item(sect, offset) {
        if (!sect) {
            // Don't include null sections
            return offset;
        }
        sect._offset = offset;
        this.sect.sect.items.push({ sect: sect.ident, offset: offset });
        return offset + sect.size;
    }
    emitSection(file, comp, sect) {
        if (sect) {
            const buf = comp.toBuffer(sect);
            if (buf.length > sect.size) {
                // buf.length may be < sect.size if there is a variable part (i.e. elem)
                throw new RangeError(`Internal Error: Section ${constants.str_section_id(sect.ident)} claimed size ${sect.size} but produced buffer of size ${buf.length}.`);
            }
            file.set(buf, sect._offset);
        }
    }
    emitStrings(file) {
        for (const item of this.sect.strs.items) {
            if (item._value === '') {
                // We have a special case for the zero-length string
                const sbuf = r.uint16le;
                file.set(sbuf.toBuffer(0), item.offset + this.sect.strs._offset);
            }
            else {
                const sbuf = new r.String(null, 'utf16le');
                file.set(sbuf.toBuffer(item._value), item.offset + this.sect.strs._offset);
            }
        }
    }
    emitElements(file) {
        if (this.sect.elem) {
            for (const str of this.sect.elem.strings) {
                if (str.items.length > 0) {
                    const COMP_PLUS_ELEM_ELEMENTS = new r.Array(this.file.COMP_PLUS_ELEM_ELEMENT, str.items.length);
                    file.set(COMP_PLUS_ELEM_ELEMENTS.toBuffer(str.items), str.offset + this.sect.elem._offset);
                }
            }
        }
    }
}
//# sourceMappingURL=kmx-plus-builder.js.map